unit AppCtrl;

// {$DEFINE DEUTSCH}

{
  by
  Thomas Schubbauer, Vienna (Austria)
  e-mail:   100640.1021@compuserve.com
  homepage: http://ourworld.compuserve.com/Homepages/SchubbT
  this unit is freeware
  comments welcome

  written with Delphi 3, should work with Delphi 2

  TAppInstance is a component, to check how many instances of an
  application are running, and if a program was not correctly terminated.
  All instances of an application must use the
  same control-file (ControlFile-property).
}

interface

uses Windows, SysUtils, Classes, Forms;

const
{$IFNDEF DEUTSCH}
  E_INVALID_CONTROL_FILE = 'Could not open or create "%0:s"';
  E_LOCK_TIMEOUT = 'Could not lock "%0:s"';
  E_WRITE_ERROR = 'Could not write to "%0:s"';
  E_CONTROL_FILE_CHANGE = 'ControlFile can not be changed when Active = True';
  E_USERNAME_CHANGE = 'UserName can not be changed when Active = True';
  E_TO_MANY_USERS = 'To many active users';
  E_TO_MANY_INSTANCES = 'To many active instances';
  E_INTERNAL_ERROR = 'Internal error in TAppInstance';
{$ELSE}
  E_INVALID_CONTROL_FILE = '"%0:s" kann nicht geffnet oder erzeugt werden';
  E_LOCK_TIMEOUT = '"%0:s" kann nicht gesperrt werden';
  E_WRITE_ERROR = '"%0:s" kann nicht geschrieben werden';
  E_CONTROL_FILE_CHANGE = 'ControlFile kann nicht gendert werden wenn Active = True';
  E_USERNAME_CHANGE = 'UserName kann nicht gendert werden wenn Active = True';
  E_TO_MANY_USERS = 'Zu viele aktive Benutzer';
  E_TO_MANY_INSTANCES = 'Zu viele aktive Programminstanzen';
  E_INTERNAL_ERROR = 'Interner Fehler in TAppInstance';
{$ENDIF}

  LOCK_OFFSET   = $80000000;
  LOCK_TIMEOUT  = 5.0; { sec }

type

  EAppInstanceError = class (Exception);

  TUserName = String [20];

  TInstance = record
    UserName  : TUserName;
    LoginTime : TDateTime;
    LogoutTime: TDateTime;
    Active    : Boolean;
  end;

  TInstanceEvent = procedure (var AInstance : TInstance) of object;

  TAppInstance = class (TComponent)
  protected
    Recno           : Integer;
    hControlFile    : THandle;
    lWantActivate   : Boolean;
    rInstance       : TInstance;
    FControlFile    : TFileName;
    FMaxInstances   : Integer;
    FMaxUsers       : Integer;
    FInstances      : TStrings;
    FUsers          : TStrings;
    FOnDetectAbort  : TInstanceEvent;
    FOnOpen         : TInstanceEvent;
    FOnClose        : TInstanceEvent;
    FOnMaxUsers     : TInstanceEvent;
    FOnMaxInstances : TInstanceEvent;
    function  IsActive : Boolean; virtual;
    procedure OpenControlFile; virtual;
    procedure CloseControlFile; virtual;
    procedure LockNextFreeRecord; virtual;
    procedure DetectAbort (AInstance : TInstance) ; virtual;
    function  LockRecord (nRecno : Integer) : Boolean;
    procedure UnlockRecord (nRecno : Integer);
    procedure FindInstance;
    procedure LockControlFile;
    procedure UnlockControlFile;
    procedure WriteInstance;
    procedure Loaded; override;
    procedure SetActive (Value : Boolean);
    function  GetActive : Boolean;
    procedure SetControlFile (const Value : TFileName);
    procedure SetMaxInstances (Value : Integer);
    procedure SetMaxUsers (Value : Integer);
    procedure SetUserName (const Value : TUserName);
  public
    constructor Create (AOwner : TComponent); override;
    destructor  Destroy; override;
    procedure Open;
    procedure Close;
    procedure Refresh; virtual;
    property  Instances     : TStrings
      read    FInstances;
    property  Users         : TStrings
      read    FUsers;
  published
    property  Active       : Boolean
      read    GetActive
      write   SetActive
      default False;
    property  ControlFile  : TFileName  { Unique file for each Application }
      read    FControlFile              { in the network }
      write   SetControlFile;
    property  MaxInstances : Integer    { max. count of TAppInstance-Objects }
      read    FMaxInstances             { with the same control-file }
      write   SetMaxInstances
      default 0;
    property  MaxUsers     : Integer    { max. count of different users of }
      read    FMaxUsers                 { a TAppInstance-Object with the }
      write   SetMaxUsers               { same control-file }
      default 0;
    property  UserName      : TUserName { net - username, if left blank }
      read    rInstance.UserName        { the windows-username is used when }
      write   SetUserName;              { Active is set to true }
    property  OnDetectAbort : TInstanceEvent
      read    FOnDetectAbort  { fired if an instance was not }
      write   FOnDetectAbort; { closed correctly }
    property  OnOpen        : TInstanceEvent
      read    FOnOpen
      write   FOnOpen;
    property  OnClose       : TInstanceEvent
      read    FOnClose
      write   FOnClose;
    property  OnMaxUsers    : TInstanceEvent
      read    FOnMaxUsers     { fired if the count of users exeeds MaxUsers}
      write   FOnMaxUsers;    { if not assigned an exeption is raised }
    property  OnMaxInstances: TInstanceEvent
      read    FOnMaxInstances { fired if the count of instances exeeds }
      write   FOnMaxInstances;{ MaxInstances,
                              { if not assigned an exeption is raised }
  end;

procedure Register;

implementation

function WinUserName : String;
var
	len : Integer;
begin
	len := 0;
  WNetGetUser (nil, nil, len);
  SetLength (Result, len);
  if WNetGetUser (nil, PChar (Result), len) = 0 then
	  Result := StrPas (PChar (Result))
  else
  	Result := '';
end;

constructor TAppInstance.Create (AOwner : TComponent);
begin
  inherited Create (AOwner);
  hControlFile  :=  INVALID_HANDLE_VALUE;
  FInstances    := TStringList.Create;
  TStringList (FInstances).Sorted := True;
  FUsers        := TStringList.Create;
  TStringList (FUsers).Sorted := True;
end;

destructor TAppInstance.Destroy;
begin
  Active := False;
  FUsers.Free;
  FInstances.Free;	
  inherited Destroy;
end;

procedure TAppInstance.Open;
begin
  Active := True;
end;

procedure TAppInstance.Close;
begin
  Active := False;
end;

procedure TAppInstance.Refresh;
var
  Instance  : TInstance;
  nRecno    : Integer;
  User      : TUserName;
begin
  if IsActive then begin
    FInstances.Clear;
    FUsers.Clear;
    FileSeek (hControlFile, 0, 0);
    nRecno := 0;
    while FileRead (hControlFile, Instance, SizeOf (Instance)) = SizeOf (Instance) do begin
      Inc (nRecno);
      if LockRecord (nRecno) then begin
        with Instance do begin
          if Active then begin
            DetectAbort (Instance);
            Active := False;
            FileSeek (hControlFile, -SizeOf (Instance), 1);
            FileWrite (hControlFile, Instance, SizeOf (Instance));
          end;
        end;
        UnLockRecord (nRecno)
      end else begin
        with Instance do begin
          FInstances.Add (UserName + ' : ' + DateTimeToStr (LoginTime));
          User := AnsiUpperCase (UserName);
          with FUsers do if IndexOf (User) < 0 then Add (User);
        end;
      end;
    end;
    if FMaxUsers > 0 then if FUsers.Count > FMaxUsers then
      if Assigned (FOnMaxUsers) then
        FOnMaxUsers (rInstance)
      else
        raise EAppInstanceError.Create (E_TO_MANY_USERS);
    if FMaxInstances > 0 then if FInstances.Count > FMaxInstances then
      if Assigned (FOnMaxInstances) then
        FOnMaxInstances (rInstance)
      else
        raise EAppInstanceError.Create (E_TO_MANY_INSTANCES);
  end;
end;

procedure TAppInstance.SetActive (Value : Boolean);
begin
  if csReading in ComponentState then
    lWantActivate := True
  else if Active <> Value then begin
    if Value then begin
      with rInstance do begin
        if (UserName = '') and not (csDesigning in ComponentState) then
          UserName := WinUserName;
        Active    := True;
        LoginTime := Now;
      end;
      if Assigned (FOnOpen) then FOnOpen (rInstance);
      OpenControlFile;
      LockNextFreeRecord;
      WriteInstance;
      Refresh;
    end else begin
      with rInstance do begin
        Active := False;
        LogoutTime := Now;
      end;
      if Assigned (FOnClose) then FOnClose (rInstance);
      WriteInstance;
      UnlockRecord (Recno);
      CloseControlFile;
      FInstances.Clear;
      FUsers.Clear; 
    end;
  end;
end;

function TAppInstance.GetActive : Boolean;
begin
  Result := (hControlFile <> INVALID_HANDLE_VALUE);
end;

procedure TAppInstance.SetControlFile (const Value : TFileName);
begin
  if Value <> FControlFile then begin
    if not Active then
      FControlFile := Value
    else
      raise EAppInstanceError.Create (E_CONTROL_FILE_CHANGE);
  end;
end;

procedure TAppInstance.SetMaxInstances (Value : Integer);
begin
  if Value <> FMaxInstances then  begin
    FMaxInstances := Value;
    Refresh;
  end;
end;

procedure TAppInstance.SetMaxUsers (Value : Integer);
begin
  if Value <> FMaxUsers then  begin
    FMaxUsers := Value;
    Refresh;
  end;
end;

procedure TAppInstance.SetUserName (const Value : TUserName);
begin
  with rInstance do begin
    if Value <> UserName then begin
      if not Active then
        UserName := Value
      else
        raise EAppInstanceError.Create (E_USERNAME_CHANGE);
    end;
  end;
end;

function TAppInstance.IsActive;
begin
  Result := Active and ([csDesigning, csReading] * ComponentState = []);
end;

procedure TAppInstance.OpenControlFile;
begin
  hControlFile :=
    CreateFile (
      PChar (FControlFile),
      GENERIC_READ or GENERIC_WRITE,
      FILE_SHARE_READ or FILE_SHARE_WRITE,
      nil,
      OPEN_ALWAYS,
      FILE_ATTRIBUTE_NORMAL,
      INVALID_HANDLE_VALUE
    );
  if hControlFile = INVALID_HANDLE_VALUE then
    raise EAppInstanceError.CreateFmt (E_INVALID_CONTROL_FILE, [FControlFile]);
end;

procedure TAppInstance.CloseControlFile;
begin
  CloseHandle (hControlFile);
  hControlFile := INVALID_HANDLE_VALUE;
  { delete, to keep ControlFile small, but maybe not all Aborts are detected }
  // DeleteFile (FControlFile);
end;

procedure TAppInstance.LockNextFreeRecord;
var
  Locked, NewRecord : Boolean;
  Instance : TInstance;
begin
  FileSeek (hControlFile, 0, 0);
  Recno     := 0;
  NewRecord := False;
  Locked    := False;
  repeat
    if FileRead (hControlFile, Instance, Sizeof (Instance)) = Sizeof (Instance) then begin
      Inc (Recno);
      Locked := LockRecord (Recno);
    end else begin
      NewRecord := True;
    end;
  until Locked or NewRecord;
  if NewRecord then begin
    LockControlFile;
    Recno := (FileSeek (hControlFile, 0, 2) + 1) div Sizeof (Instance) + 1;
    if not LockRecord (Recno) then
      raise EAppInstanceError.Create (E_INTERNAL_ERROR);
  end else begin
    if Instance.Active then DetectAbort (Instance);
  end;
  WriteInstance;
  if NewRecord then UnlockControlFile;
end;

function TAppInstance.LockRecord (nRecno : Integer) : Boolean;
begin
  Result := LockFile (hControlFile, LOCK_OFFSET + nRecno, 0, 1, 0);
end;

procedure TAppInstance.UnlockRecord (nRecno : Integer);
begin
  UnLockFile (hControlFile, LOCK_OFFSET + nRecno, 0, 1, 0);
end;

procedure TAppInstance.LockControlFile;
var
  TimeOut : TDateTime;
begin
  TimeOut := Now + LOCK_TIMEOUT / 24 / 60 / 60;
  while not LockRecord (0) do begin
    Application.ProcessMessages;
    if Now > TimeOut then
      raise EAppInstanceError.CreateFmt (E_LOCK_TIMEOUT, [FControlFile]);
  end;
end;

procedure TAppInstance.UnLockControlFile;
begin
  UnlockRecord (0);
end;

procedure TAppInstance.FindInstance;
begin
  FileSeek (hControlFile, (Recno - 1) * SizeOf (rInstance), 0);
end;

procedure TAppInstance.WriteInstance;
begin
  FindInstance;
  if FileWrite (hControlFile, rInstance, Sizeof (rInstance)) < Sizeof (rInstance) then
    raise EAppInstanceError.CreateFmt (E_WRITE_ERROR, [FControlFile]);
  FlushFileBuffers (hControlFile);
end;

procedure TAppInstance.DetectAbort (AInstance : TInstance);
begin
  if Assigned (FOnDetectAbort) then FOnDetectAbort (AInstance);
end;

procedure TAppInstance.Loaded;
begin
  inherited Loaded;
  if lWantActivate then Active := True;
end;

procedure Register;
begin
  RegisterComponents ('AppUtil', [TAppInstance]);
end;

end.
